#
# Compute AUC and mAP for localization + Accuracy and mAP for classification.
#
# Pascal Mettes (2019).
# Recognizing Unseen Actions from Objects.
#

import os
import sys
import h5py
import numpy as np
import argparse
import configparser
from   sklearn.metrics import average_precision_score

import helper

#
# Perform score normalization.
#
def norm_scores(scores):
    #amin, amax = np.min(scores), np.max(scores)
    #scores = (scores - amin) / (amax - amin)
    #for i in xrange(scores.shape[1]):
    #    amin, amax = np.min(scores[:,i,:]), np.max(scores[:,i,:])
    #    scores[:,i,:] = (scores[:,i,:] - amin) / (amax - amin)
    for i in range(scores.shape[0]):
        amin, amax = np.min(scores[i,:,:]), np.max(scores[i,:,:])
        scores[i,:,:] = (scores[i,:,:] - amin) / (amax - amin)
    return scores

#
# Parse user arguments.
#
def parse_args():
    parser = argparse.ArgumentParser(description="Zero-shot evaluation")
    parser.add_argument("-c", dest="configfile", default="", type=str)
    parser.add_argument("-t", dest="tubedir", default="", type=str)
    #parser.add_argument("-i", dest="ioudir", default="", type=str)
    parser.add_argument("-v", dest="videoscores", default="", type=str)
    parser.add_argument("-n", dest="nr_tubes", default=1, type=int)
    parser.add_argument("-m", dest="mode", default="local", type=str)
    args = parser.parse_args()
    return args

#
# Entry point of the script.
#
if __name__ == "__main__":
    # Parse user arguments.
    args = parse_args()
    
    # Parse the configurationfile.
    parser = configparser.ConfigParser()
    parser.read(args.configfile)
    # Yield the appropriate directories.
    gtdir     = parser.get('actions', 'gtdir')
    split     = parser.get('actions', 'testsplit')
    vidlist   = parser.get('actions', 'vidfile')
    ioudir    = args.tubedir.replace("/tubes/", "/iou/")
    dataset   = args.configfile.split("/")[1].split(".")[0]
    
    # Load the split indices and set the training videos.
    teidxs = np.loadtxt(split, dtype=int) - 1
    videos = [line.strip().split()[0] for line in open(vidlist)]
    videos = np.array(videos)
    videos = videos[teidxs]
    # Yield action names.
    actions = np.array([vid.split("/")[0] for vid in videos])
    actions, aidxs = np.unique(actions, return_inverse=True)
    
    # Initialize the score and label containers.
    overlaps = []
    labels   = []
    predictions_local  = []
    predictions_global = []
    gtmatches = []
    for i in range(len(actions)):
        overlaps.append([])
        labels.append([])
        predictions_local.append([])
        predictions_global.append([])
        gtmatches.append([])
    actioncount = np.zeros(len(actions), dtype=int)
    
    # Load video scores.
    if args.mode != "local":
        vidscores = np.load(args.videoscores)
        vidscores = vidscores[teidxs]
    
    # Go over all videos and actions.
    for i in range(len(videos)):
        print("%d/%d" %(i+1, len(videos)), end="\r"),
        sys.stdout.flush()
    
        # Load GT and update counter.
        if dataset == "h2t":
            gttube = h5py.File(gtdir + videos[i] + "/annos.hdf5")
        else:
            gttube = h5py.File(gtdir + videos[i] + "/gt/tubes.hdf5", "r")
        actioncount[aidxs[i]] += len(gttube)
    
        for j in range(len(actions)):
            # Load tubes and scores.
            tubes  = h5py.File(args.tubedir + videos[i] + \
                "/tubes-%s.hdf5" %(actions[j]), "r")
            #tubes = h5py.File(args.tubedir + actions[j] + "/" + videos[i] + "/tubes.hdf5", "r")
            # Load pre-computed overlaps.
            #computed_overlaps = np.load(ioudir + actions[j] + "/" + videos[i] + "/iou.npy")
            computed_overlaps = np.load(ioudir + videos[i] + \
                "/iou-%s.npy" %(actions[j]))
            
            # Handle all tubes accordingly.
            for k in range(computed_overlaps.shape[1]):
                nt = min(args.nr_tubes, computed_overlaps.shape[0])
                noverlaps = np.zeros(args.nr_tubes)
                nlabels   = np.zeros(args.nr_tubes, dtype=int)
                nscoreslocal  = np.zeros(args.nr_tubes) - 1e5
                nscoresglobal = np.zeros(args.nr_tubes) - 1e5

                noverlaps[:nt] = computed_overlaps[:nt,k].copy()
                overlaps[j].append(noverlaps)
                nlabels[:nt]   = j == aidxs[i]
                labels[j].append(nlabels)
                if args.mode != "local":
                    nscoresglobal[:nt] = vidscores[i,j]
                    predictions_global[j].append(nscoresglobal)
                if args.mode != "global":
                    for l in range(nt):
                        nscoreslocal[l] = np.mean(tubes["scores-%d" %(l)])
                    predictions_local[j].append(nscoreslocal)

                '''overlaps[j].append(computed_overlaps[:nt,k])
                labels[j].append([j == aidxs[i]] * nt)
                if args.mode != "local":
                    predictions_global[j].append(np.ones(nt) * vidscores[i,j])
                if args.mode != "global":
                    vpd = np.zeros(nt)
                    for l in xrange(nt):
                        vpd[l] = np.mean(tubes["scores-%d" %(l)])
                    predictions_local[j].append(vpd)'''
    print()
    
    # To array.
    overlaps = np.array(overlaps)
    labels   = np.array(labels, dtype=bool)
    predictions_local  = np.array(predictions_local)
    predictions_global = np.array(predictions_global)
    print(overlaps.shape, labels.shape, predictions_global.shape)

    # Normalize the scores.
    if args.mode != "global":
        #predictions_dom = add_negative_info(predictions_dom, alpha)
        predictions_local = norm_scores(predictions_local)
    if args.mode != "local":
        #predictions_sup = add_negative_info(predictions_sup, alpha)
        predictions_global = norm_scores(predictions_global)

    # Fuse the scores.
    if args.mode == "local":
        predictions = predictions_local
    elif args.mode == "global":
        predictions = predictions_global
    else:
        predictions = 0.05 * predictions_local + predictions_global

    # Compute localization scores.
    iout = np.arange(0.1,1.01,0.1)
    aucs = helper.compute_auc(overlaps, labels, predictions, actioncount, iout)
    aps  = helper.compute_ap(overlaps, labels, predictions, actioncount, iout)
    taps = np.mean(aps, axis=0)
    
    # Print localization results.
    print("LOCALIZATION\n\tiou:", end=", ")
    for i in range(len(iout)):
        print("%.4f," %(iout[i]), end=", ")
    print("\n\tauc:", end=", ")
    for i in range(len(iout)):
        print("%.4f," %(aucs[i]), end=", ")
    print("\n\tmap:", end=", ")
    for i in range(len(iout)):
        print("%.4f," %(taps[i]), end=", ")
    print("\nAP @ 0.2:", end=", ")
    for i in range(aps.shape[0]):
        print("%.4f" %(aps[i,1]), end=", ")
    print()
    print("AP @ 0.5:", end=", ")
    for i in range(aps.shape[0]):
        print("%.4f" %(aps[i,4]), end=", ")
    print()
    
    # Classification.
    # Multi-label evaluation.
    if dataset == "h2t":
        exit()
        tp  = np.max(predictions, axis=2)
        ty  = labels[:,:,0]
        uv,ui,uc = np.unique(vidnames, return_inverse=True, return_counts=True)
        aps = np.zeros(labels.shape[0])
        for i in range(labels.shape[0]):
            toremove = []
            for j in range(len(uv)):
                if uc[j] > 1:
                    vidxs = np.where(ui == j)[0]
                    ocidx = -1
                    for k in range(len(vidxs)):
                        if aidxs[vidxs[k]] == i:
                            ocidx = k
                    if ocidx < -1:
                        np.random.choice(np.arange(len(vidxs)))
                    toremove.append(np.delete(vidxs, ocidx))
            toremove = np.concatenate(toremove)
            tokeep   = np.ones(tp.shape[1], dtype=bool)
            tokeep[toremove] = 0
            atp = tp[i,tokeep]
            aty = ty[i,tokeep]
            aps[i] = average_precision_score(aty, atp)

        print("\nCLASSIFICATION")
        print("\taps:", end=" ")
        for i in range(len(aps)):
            print("%.4f," %(aps[i]), end=", ")
        print("(%.4f)" %(np.mean(aps)))
    # Standard multi-class evaluation
    else:
        tp = np.max(predictions, axis=2)
        yp = np.argmax(tp, axis=0)
        yt = aidxs
        a  = np.bincount(yt[yt == yp], minlength = len(actions)).astype(float)
        b  = np.bincount(yt).astype(float)
            
        na   = np.mean(a / b)
        ua   = np.sum(yt == yp) / float(len(yt))
        maps = np.zeros(len(actions))
        for i in range(labels.shape[0]):
            maps[i] = average_precision_score(labels[i,:,0], tp[i,:])
    
        # Print classification results.
        print("\nCLASSIFICATION")
        print("\taccuracy: %.4f (normalized), %.4f (unnormalized)" %(na, ua))
        print("\taps:", end=" ")
        for i in range(len(maps)):
            print("%.4f," %(maps[i]), end=", ")
        print("(%.4f)" %(np.mean(maps)))
